package net.indiespot.continuations;

import java.util.AbstractQueue;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.PriorityQueue;

class VirtualThreadQueue extends AbstractQueue<VirtualThread> {
	private final VirtualProcessor proc;
	private final Comparator<VirtualThread> comp;
	private final PriorityQueue<VirtualThread> head;
	private ArrayList<VirtualThread>[] buckets;
	private ArrayList<VirtualThread> tail, tailNext;

	private long bucketDuration;
	private long headTime = Long.MIN_VALUE;

	public VirtualThreadQueue(VirtualProcessor proc, long bucketDuration) {
		this.proc = proc;
		this.bucketDuration = bucketDuration;
		this.comp = new VirtualProcessor.VirtualThreadComparator();

		head = new PriorityQueue<>(11, comp);
		buckets = new ArrayList[4];
		for (int i = 0; i < buckets.length; i++) {
			buckets[i] = new ArrayList<>();
		}
		tail = new ArrayList<>();
		tailNext = new ArrayList<>();
	}

	private void addBuckets() {
		int prevBucketCount = buckets.length;
		buckets = Arrays.copyOf(buckets, buckets.length * 4 / 3);
		for (int i = prevBucketCount; i < buckets.length; i++) {
			buckets[i] = new ArrayList<>();
		}

		// redistribute tail elements over new buckets (and tail)
		List<VirtualThread> copy = tail;
		tail = new ArrayList<>();
		this.addAll(copy);
	}

	public boolean isEmpty() {
		for (int i = 0; i < buckets.length; i++) {
			if (!buckets[i].isEmpty()) {
				return false;
			}
		}
		return head.isEmpty() && tail.isEmpty();
	}

	@Override
	public VirtualThread peek() {
		this.sync();

		return head.peek();
	}

	@Override
	public VirtualThread poll() {
		this.sync();

		return head.poll();
	}

	@Override
	public boolean offer(VirtualThread thread) {
		this.sync();

		// insert VirtualThread into head, bucket or tail
		long toWait = thread.wakeUpAt - headTime;
		int bucketIndex = (int) (toWait / bucketDuration);
		if (bucketIndex <= 0) {
			head.add(thread);
		} else {
			bucketIndex -= 1;

			if (bucketIndex < buckets.length) {
				buckets[bucketIndex].add(thread);
			} else {
				tail.add(thread);

				if (this.isTailSplitNeeded()) {
					this.addBuckets();
				}
			}
		}

		return true;
	}

	@Override
	public boolean remove(Object obj) {
		VirtualThread thread = (VirtualThread) obj;

		this.sync();

		// remove VirtualThread from head, bucket or tail
		long toWait = ((VirtualThread) thread).wakeUpAt - headTime;
		int bucketIndex = (int) (toWait / bucketDuration);
		if (bucketIndex <= 0) {
			return head.remove(thread);
		}

		bucketIndex -= 1;
		if (bucketIndex < buckets.length) {
			return buckets[bucketIndex].remove(thread);
		}
		return tail.remove(thread);
	}

	private boolean isTailSplitNeeded() {
		return buckets.length < 10_000 && tail.size() > 256;
	}

	private long sync() {
		long procTime = proc.getCurrentTime();
		if (headTime == Long.MIN_VALUE) {
			headTime = procTime;
		}

		while (procTime >= headTime + bucketDuration) {
			headTime += bucketDuration;

			// move first bucket into 'head'
			ArrayList<VirtualThread> first = buckets[0];
			first.trimToSize(); // shrink to typical size
			for (VirtualThread elem : first) {
				head.add(elem);
			}
			first.clear();
			ArrayList<VirtualThread> last = first;

			// shift all buckets one index, set 'first' as last bucket
			System.arraycopy(buckets, 1, buckets, 0, buckets.length - 1);
			buckets[buckets.length - 1] = last;

			// split 'tail' into new tail and last bucket
			long tailThreshold = headTime + bucketDuration * (buckets.length + 1);
			tail.trimToSize(); // shrink to typical size
			for (VirtualThread elem : tail) {
				if (elem.wakeUpAt < tailThreshold) {
					last.add(elem);
				} else {
					tailNext.add(elem);
				}
			}
			tail.clear();

			// swap
			ArrayList<VirtualThread> temp = tail;
			tail = tailNext;
			tailNext = temp;

		}
		return procTime;
	}

	@Override
	public int size() {
		int sum = head.size() + tail.size();
		for (int i = 0; i < buckets.length; i++) {
			sum += buckets[i].size();
		}
		return sum;
	}

	@Override
	public boolean contains(Object o) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean containsAll(Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	@Override
	public Iterator<VirtualThread> iterator() {
		throw new UnsupportedOperationException();
	}
}
